import logging
from typing import Sequence, Optional, List, Callable, Any, Tuple
import networkx as nx
from .step import Step
from .stepfunctions.steps import LambdaStep, Chain, Retry, Parallel, Graph

logger = logging.getLogger(__name__)


def crawl_back(graph: nx.DiGraph, step: Step):
    """Create the Graph of Steps

    Args:
        graph (DiGraph): The graph to populate with Steps
        step (Step): The `Step` to add to the graph
    """
    for dependency in step.depends_on:
        graph.add_edge(dependency, step)
        crawl_back(graph, dependency)


def convert_step_to_lambda(
    step: Step, generate_step_name: Callable[[Step], str]
) -> LambdaStep:
    """Create Lambda from Step

    Args:
        step (Step): The `Step` to convert to a lambda
        generate_step_name (callable): Generates the ARN of the Lambda from the step
    """
    lambda_state = LambdaStep(
        state_id=step.name,
        parameters={
            "FunctionName": generate_step_name(step),  # the function arn
            "Payload.$": "$",  # pass in all the possible values, including outputs from previous steps
        },
    )
    if step.retry_count > 0:
        lambda_state.add_retry(
            Retry(
                error_equals=["States.TaskFailed"],
                interval_seconds=15,
                max_attempts=step.retry_count,
                backoff_rate=4.0,
            )
        )
    logger.debug(f"Converted {step.name} to Lambda")
    return lambda_state


def _default_lambda_name(s: Step) -> str:
    return "${aws_lambda_function." + s.name + "lambda.arn}"


class Pipeline:
    def __init__(
        self,
        name: str = "",
        steps: Optional[Sequence[Step]] = None,
        schedule: Optional[str] = None,  # cron
        generate_step_name: Callable[[Step], str] = _default_lambda_name,
    ):
        """Initialize a Pipeline

        Args:
            name (str): The name of the pipeline.
            steps (Sequence[Step]): The list of the non-conditional Steps associated with the pipeline.
        """
        self.name = name
        self.steps = steps if steps else []
        self.graph = nx.DiGraph()
        self.generate_step_name = generate_step_name
        self.schedule = schedule
        for step in steps:
            crawl_back(self.graph, step)
        if not nx.is_directed_acyclic_graph(self.graph):
            raise ValueError("Cycle detected in pipeline step graph.")
        if not self._check_uniqueness_of_names():
            raise ValueError(
                "Non-unique Step names!  Step names must be unique or a unique name must be passed to the @step decorator."
            )

    def _check_uniqueness_of_names(self) -> bool:
        num_steps = len(self.get_steps())
        return len(set(step.name for step in self.get_steps())) == num_steps

    def get_steps(self) -> List[Step]:
        """Gets all steps, guaranteed to be unique."""
        return self.graph.nodes

    def generate_layers(self) -> List[List[Step]]:
        """
        Create indexed sets of steps.
        This allows steps to be run in parallel,
        if they don't depend on each other
        """
        return list(nx.topological_generations(self.graph))

    def generate_step_functions(self) -> dict:
        """Create Step Function workflow definition"""
        dag_lambda = []
        for index, layer in enumerate(self.generate_layers()):
            if len(layer) == 1:
                dag_lambda.append(
                    convert_step_to_lambda(layer[0], self.generate_step_name)
                )
            else:
                parallel_state = Parallel(f"parallel at {index}")
                for step in layer:
                    parallel_state.add_branch(
                        convert_step_to_lambda(step, self.generate_step_name)
                    )
                dag_lambda.append(parallel_state)
        chain = Chain(dag_lambda)
        workflow = Graph(
            chain
        )  # Workflow(name=self.name, definition=chain, role="doesnotmatter")
        logger.debug(f"Converted {self.name} to step function Workflow")
        return workflow.to_dict()

    def set_generate_step_name(self, generate_step_name: Callable[[Step], str]):
        self.generate_step_name = generate_step_name

    def local_run(self) -> List[List[Tuple[str, Any]]]:
        """
        Runs pipeline locally, with no AWS dependency.
        Returns all intermediary outputs.
        """
        outputs = {}  # contains all intermediary output
        output_arr = (
            []
        )  # contains all intermediary output, in the shape of the steps given by the topological generations
        for layer in self.generate_layers():
            layer_output = []
            for step in layer:
                args = []
                for arg in step.args:
                    if isinstance(arg, Step):
                        args.append(outputs[arg.name])
                    else:
                        args.append(arg)
                outputs[step.name] = step.func(*args)
                logger.debug(f"Output from step {step.name}: {outputs[step.name]}")
                layer_output.append((step.name, outputs[step.name]))
            output_arr.append(layer_output)
        return output_arr



import json
import collections


def flatten(input, parent_key="", sep="."):
    items = []
    for k, v in input.items():
        if parent_key:
            flattened_key = parent_key + sep + k
        else:
            flattened_key = k
        if isinstance(v, dict):
            items.extend(flatten(v, flattened_key, sep=sep).items())
        else:
            items.append((flattened_key, v))
    return dict(items)


def replace_type_with_str(schema):
    schema_with_str = {}
    for k, v in schema.items():
        if isinstance(v, dict):
            schema_with_str[k] = replace_type_with_str(v)
        else:
            schema_with_str[k] = v.__name__
    return schema_with_str


ValidationResult = collections.namedtuple(
    "ValidationResult", "valid keys_missing keys_type_mismatch"
)


class Placeholder(object):
    """
    A collection of Placeholder variables.
    """

    def __init__(self, schema=None, **kwargs):
        """
        Args:
            schema (dict, optional): Schema for the placeholder collection. (default: None)
                                     Example below::
                                        {
                                            'ModelName': str,
                                            'JobName': str,
                                            'Hyperparameters': {
                                                'tol': float
                                            }
                                        }

        Keyword Args:
            name (str, optional): Name of the placeholder variable. (default: None)
            type (type, optional): Type of the placeholder variable. (default: None)
            parent (Placeholder, optional): Parent variable for a placeholder variable. (default: None)
        """
        self.store = {}
        self.immutable = False
        self.schema = schema
        if self.schema:
            self._set_schema(schema)
            self._make_immutable()
        self.json_str_template = "{}"

        self.name = kwargs.get("name")
        self.type = kwargs.get("type")
        self.parent = kwargs.get("parent")

    def get(self, name, type):
        """
        Create a placeholder variable with an associated type.

        Args:
            name (str): Name of the placeholder variable.
            type (type): Type of the placeholder variable.

        Raises:
            ValueError: If placeholder variable with the same name but different type already exists.
            ValueError: If placeholder variable does not fit into a previously specified schema for the placeholder collection.

        Returns:
            Placeholder: Placeholder variable.
        """
        if not self._is_valid_name(name):
            raise ValueError("Key name can only be string or integer")
        if name in self.store:
            curr_variable = self.store[name]
            if curr_variable.type != type:
                raise ValueError(
                    "Key already exists with a different value type: {current_value_type}".format(
                        current_value_type=curr_variable.type
                    )
                )
            return curr_variable
        else:
            self.store[name] = self._create_variable(name=name, parent=self, type=type)
            return self.store[name]

    def get_schema_as_dict(self):
        """
        Generate a schema for the placeholder collection as a Python dictionary.

        Returns:
            dict: Placeholder collection schema.
        """
        schema = {}
        for k, v in self.store.items():
            if v._is_empty():
                schema[k] = v.type or str
            else:
                schema[k] = v.get_schema_as_dict()
        return schema

    def get_schema_as_json(self, pretty=False):
        """
        Generate a schema for the placeholder collection as a JSON formatted string.

        Args:
            pretty (bool, optional): Boolean flag set to `True` if JSON string should be prettified. `False`, otherwise. (default: False)

        Returns:
            str: JSON formatted string representation of the block.
        """
        dict_schema_str = replace_type_with_str(self.get_schema_as_dict())

        if pretty:
            return json.dumps(dict_schema_str, indent=4)

        return json.dumps(dict_schema_str)

    def contains(self, placeholder):
        """
        Check if the placeholder collection contains the specified placeholder variable.

        Args:
            placeholder (Placeholder): Placeholder variable to search for, in the collection.

        Returns:
            bool: `True` if placeholder variable was found in the collection. `False`, otherwise.
        """
        for k, v in self.store.items():
            if placeholder == v:
                return True
            elif v.contains(placeholder):
                return True
        return False

    def __contains__(self, placeholder):
        """
        Containment check operator for placeholder variables.
        """
        return self.contains(placeholder)

    def validate(self, input):
        """
        Validate a specified input against the placeholder collection schema.

        Args:
            input (dict): Input to validate against the placeholder collection schema.

        Returns:
            ValidationResult: Named tuple with the keys:
                                `valid` (Boolean): Representing the result of validation ,
                                `keys_missing` (list(str)): List of keys missing in the input ,
                                `keys_type_mismatch` (list(str), type, type): List of tuples with key name, expected type, and provided type.
        """
        if input is None:
            return False, None, None
        flattened_schema = flatten(self.get_schema_as_dict())
        flattened_input = flatten(input)
        keys_missing = [i for i in flattened_schema if i not in flattened_input]
        keys_type_mismatch = []
        for k, v in flattened_input.items():
            if k in flattened_schema and not isinstance(v, flattened_schema.get(k)):
                keys_type_mismatch.append((k, flattened_schema.get(k), type(v)))
        if len(keys_missing) > 0 or len(keys_type_mismatch) > 0:
            valid = False
        else:
            valid = True
        return ValidationResult(
            valid=valid,
            keys_missing=keys_missing,
            keys_type_mismatch=keys_type_mismatch,
        )

    def _create_variable(self, name, parent, type=None):
        raise NotImplementedError

    def _get_path(self):
        """
        Get path to a placeholder variable node in the collection.
        """
        path = []
        node = self
        while node.name is not None:
            path.append(node.name)
            node = node.parent
        path.reverse()
        return path

    def _is_empty(self):
        """
        Check if the store for a placeholder collection/variable is empty.
        """
        return len(self.store) == 0

    def _set_schema(self, schema, path=[]):
        """
        Set the schema for a placeholder collection.
        """
        for k, v in schema.items():
            if isinstance(v, dict):
                self._set_schema(v, path + [k])
            else:
                current = self
                for node in path:
                    current = current.get(node, dict)
                temp = current.get(k, v)

    def _make_immutable(self):
        """
        Make a placeholder collection (including all variables contained) immutable.
        """
        for k, v in self.store.items():
            if isinstance(v, Placeholder):
                v._make_immutable()
        self.immutable = True

    def _is_valid_name(self, name):
        if isinstance(name, str) or isinstance(name, int):
            return True
        else:
            return False

    def __getitem__(self, name):
        """
        Subscript operator to build placeholder variables.
        """
        if not self._is_valid_name(name):
            raise ValueError("Key name can only be string or integer")
        if name in self.store:
            return self.store[name]
        else:
            self.store[name] = self._create_variable(name=name, parent=self)
            return self.store[name]

    def _join_path(self, path):
        subscript_list = []
        for i in path:
            if isinstance(i, str):
                subscript_list.append("['{}']".format(i))
            elif isinstance(i, int):
                subscript_list.append("[{}]".format(i))
        return "".join(subscript_list)

    def to_jsonpath(self):
        """
        Returns a JSON path representation of the placeholder variable to be used for step parameters.

        Returns:
            str: JSON path representation of the placeholder variable
        """
        return self.json_str_template.format(self._join_path(self._get_path()))


class StepInput(Placeholder):
    """
    Top-level class for step input placeholders.
    """

    def __init__(self, schema=None, **kwargs):
        super(StepInput, self).__init__(schema, **kwargs)
        self.json_str_template = "${}"

    def _create_variable(self, name, parent, type=None):
        """
        Creates a placeholder variable for Step Input.
        A placeholder variable can only be created if the collection is not immutable due to a pre-specified schema.
        """
        if self.immutable:
            raise ValueError(
                "Placeholder variable does not conform to schema set for the placeholder collection."
            )
        if type:
            return StepInput(name=name, parent=parent, type=type)
        else:
            return StepInput(name=name, parent=parent)



import pickle

"{{PUT_FUNCTION_HERE}}"


def combine_payload(event):
    """Takes payload from event.  If previous "step" was a Parallel state,
            this will be an array of payloads from however many steps
            were in the Parallel state.  In this case, it combines these
            outputs into one large payload.
    Args:
        event (dict): object passed to the Lambda
    """
    payload = {}
    if isinstance(event, dict):
        if "Payload" in event:
            payload = event["Payload"]
    else:  # then event is a list, and contains "multiple" payloads
        for ev in event:
            if "Payload" in ev:
                payload = {**payload, **ev["Payload"]}
    return payload


# Retrieve transform job name from event and return transform job status.
def lambda_handler(event, context):

    with open("args.pickle", "rb") as f:
        args = pickle.load(f)

    with open("name.pickle", "rb") as f:
        name = pickle.load(f)

    arg_values = []
    payload = combine_payload(event)
    for arg in args:
        if arg in payload:
            # extract the output from a previous Lambda
            arg_values.append(payload[arg])
        else:
            # just use the hardcoded argument
            arg_values.append(arg)

    result = "{{PUT_FUNCTION_NAME_HERE}}"(*arg_values)
    ## all outputs from all lambdas are stored in the payload and
    ## passed on to the next lambda(s) in the step.  This mirrors
    ## the local_run from the `Pipeline` class.  On each subsequent
    ## "step" this payload will grow larger.  At the final step, this
    ## will include the output of all intermediary steps.
    return {name: result, **payload}



from enum import Enum


class Field(Enum):

    # Common fields
    Comment = "comment"
    InputPath = "input_path"
    OutputPath = "output_path"
    Parameters = "parameters"
    ResultPath = "result_path"
    Next = "next"
    Retry = "retry"
    Catch = "catch"
    Branches = "branches"
    End = "end"
    Version = "version"

    # Pass state fields
    Result = "result"

    # Fail state fields
    Error = "error"
    Cause = "cause"

    # Wait state fields
    Seconds = "seconds"
    Timestamp = "timestamp"
    SecondsPath = "seconds_path"
    TimestampPath = "timestamp_path"

    # Choice state fields
    Choices = "choices"
    Default = "default"

    # Map state fields
    Iterator = "iterator"
    ItemsPath = "items_path"
    MaxConcurrency = "max_concurrency"

    # Task state fields
    Resource = "resource"
    TimeoutSeconds = "timeout_seconds"
    TimeoutSecondsPath = "timeout_seconds_path"
    HeartbeatSeconds = "heartbeat_seconds"
    HeartbeatSecondsPath = "heartbeat_seconds_path"

    # Retry and catch fields
    ErrorEquals = "error_equals"
    IntervalSeconds = "interval_seconds"
    MaxAttempts = "max_attempts"
    BackoffRate = "backoff_rate"
    NextStep = "next_step"



from constructs import Construct
from cdktf import App, TerraformStack, IResolvable, TerraformOutput
from cdktf_cdktf_provider_aws.provider import AwsProvider
from cdktf_cdktf_provider_aws import (
    data_aws_subnets,
    security_group,
    iam_role_policy_attachment,
    lambda_function,
    sfn_state_machine,
    iam_role,
    iam_policy,
    cloudwatch_event_rule,
    cloudwatch_event_target,
    cloudwatch_log_group,
)
from importlib import resources as impresources
from .step import Step, step
from .pipeline import Pipeline
from typing import List, Union, Optional, Tuple
import json
import zipfile
import pickle
from pathlib import Path
import inspect
import textwrap
from hashlib import sha256
import logging
import os

logger = logging.getLogger(__name__)


def remove_decorators(src: str) -> str:
    """Removes any decorators from source code and
            truncates code so that there is no space
            in front of function definition

    Args:
        src (str): Source code for lambda
    """
    lines = []
    for line in src.splitlines():
        if not line.lstrip().startswith("@"):
            lines.append(line)
    return textwrap.dedent("\n".join(lines))


def get_python_code(python_template_path: str, step: Step) -> str:
    """Generates full python code for lambda

    Args:
        python_template_path (str): Location of template python code
        step (Step): Step to place inside template
    """
    code = remove_decorators(inspect.getsource(step.func))
    logger.debug(f"Code for {step.name}: {code}")
    with open(python_template_path, "r") as f:
        template = f.read()
    template = template.replace('"{{PUT_FUNCTION_HERE}}"', code)
    template = template.replace('"{{PUT_FUNCTION_NAME_HERE}}"', step.func.__name__)
    new_file_name = f"{step.func.__name__}.py"
    with open(new_file_name, "w") as f:
        f.write(template)
    return new_file_name


def package_lambda(
    python_template_path: str, step: Step, lambda_entry: str
) -> Tuple[str, str]:
    """Creates zip of `Step` code for use in Lambda

    Args:
        python_template_path (str): Location of template python code
        step (Step): `Step` to place inside template
        lambda_entry (str): Name of python entry file
    """
    lambda_python_file = get_python_code(python_template_path, step)
    args_file = "args.pickle"
    name_file = "name.pickle"

    with open(args_file, "wb") as f:
        pickle.dump(
            [step.name if isinstance(step, Step) else step for step in step.args], f
        )

    with open(name_file, "wb") as f:
        pickle.dump(step.name, f)

    zip_name = f"{step.name}.zip"

    zf = zipfile.ZipFile(zip_name, mode="w")
    zf.write(lambda_python_file, arcname=f"{lambda_entry}.py")
    zf.write(args_file)
    zf.write(name_file)
    zf.close()
    with open(zip_name, "rb") as f:
        data = f.read()
        hash_sha256 = sha256(data).hexdigest()
    os.remove(args_file)
    os.remove(name_file)
    os.remove(lambda_python_file)
    logger.info(f"Successfully packaged files for Lambda {step.name}")

    return zip_name, hash_sha256


def generate_lambda_function(
    scope: Construct,
    name_prefix: str,
    step: Step,
    template_file: str,
    subnet_ids: Optional[List[str]] = None,
    security_group_ids: Optional[List[str]] = None,
):
    """Creates Terraform resource for Lambda.  Automatically
        adds an environment variable "VAULT_LAMBDA_ROLE" for
        easier Vault integration.

    Args:
        scope
        name_prefix (str): Prefix for lambda name to ensure uniqueness.
        step (Step): Step to create Lambda from
        template_file (str): Location of template file to populate.  Defaults to internal template, but a custom file can be provided.
        subnet_ids (list): Optional subnet IDs.  Required if VPC is specified.
        security_group_ids (list): Option security group IDs.  Required if VPC is specified.
    """
    role = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": "sts:AssumeRole",
                "Principal": {"Service": ["lambda.amazonaws.com"]},
            }
        ],
    }
    policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": [
                    "logs:CreateLogGroup",
                    "logs:CreateLogStream",
                    "logs:PutLogEvents",
                ],
                "Resource": ["arn:aws:logs:*:*:*"],
            },
        ],
    }

    lambda_entry = "index"

    lambda_filename, sha256_hash = package_lambda(template_file, step, lambda_entry)
    lambda_handler = f"{lambda_entry}.lambda_handler"

    lambda_role = iam_role.IamRole(
        scope,
        f"{step.name}role",
        assume_role_policy=json.dumps(role),
        name_prefix=step.name,
    )
    cloudwatch_policy = iam_policy.IamPolicy(
        scope,
        f"{step.name}policy",
        policy=json.dumps(policy),
    )
    lambda_policy_attachment = iam_role_policy_attachment.IamRolePolicyAttachment(
        scope,
        f"{step.name}policyattachment",
        policy_arn=cloudwatch_policy.arn,
        role=lambda_role.name,
    )
    for index, additional_policy in enumerate(step.additional_policies):
        local_policy = iam_policy.IamPolicy(
            scope,
            f"{step.name}_{index}_policy",
            policy=additional_policy,
        )
        local_policy_attachment = iam_role_policy_attachment.IamRolePolicyAttachment(
            scope,
            f"{step.name}_{index}policyattachment",
            policy_arn=local_policy.arn,
            role=lambda_role.name,
        )
    vpc_config = (
        None
        if subnet_ids is None
        else {"subnet_ids": subnet_ids, "security_group_ids": security_group_ids}
    )
    TerraformOutput(scope, f"{step.name}_lambda_role_arn", value=lambda_role.arn)
    lambda_f = lambda_function.LambdaFunction(
        scope,
        step.name,
        function_name=f"{name_prefix}_{step.name}",
        role=lambda_role.arn,
        filename=lambda_filename,
        timeout=900,
        memory_size=step.memory_size,
        runtime=step.python_runtime,
        handler=lambda_handler,
        vpc_config=vpc_config,
        layers=step.layers,
        source_code_hash=sha256_hash,
        environment={
            "variables": {**step.env_variables, "VAULT_AUTH_ROLE": lambda_role.name}
        },
    )
    TerraformOutput(scope, f"{step.name}_lambda_arn", value=lambda_f.arn)
    return lambda_f


def generate_event_bridge(scope: Construct, pipeline: Pipeline, step_function_arn: str):
    """Creates Terraform resource for event bridge to schedule step function run

    Args:
        scope
        pipeline (Pipeline): pipeline to convert into step function
        step_function_arn (str): ARN of the step function pipeline
    """
    role = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Principal": {"Service": "events.amazonaws.com"},
                "Action": "sts:AssumeRole",
            }
        ],
    }
    policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": ["states:StartExecution"],
                "Resource": step_function_arn,
            }
        ],
    }
    eventbridge_policy = iam_policy.IamPolicy(
        scope,
        f"{pipeline.name}eventbridgepolicy",
        policy=json.dumps(policy),
    )
    eventbridge_role = iam_role.IamRole(
        scope,
        f"{pipeline.name}eventbridgerole",
        assume_role_policy=json.dumps(role),
        name_prefix=pipeline.name,
    )
    eventbridge_policy_attachment = iam_role_policy_attachment.IamRolePolicyAttachment(
        scope,
        f"{pipeline.name}eventbridgepolicyattachment",
        policy_arn=eventbridge_policy.arn,
        role=eventbridge_role.name,
    )
    event_rule = cloudwatch_event_rule.CloudwatchEventRule(
        scope,
        f"{pipeline.name}eventrule",
        schedule_expression=pipeline.schedule,
    )
    cloudwatch_event_target.CloudwatchEventTarget(
        scope,
        f"{pipeline.name}eventtarget",
        rule=event_rule.name,
        arn=step_function_arn,
        role_arn=eventbridge_role.arn,
    )


def generate_step_function(
    scope: Construct, pipeline: Pipeline, aws_region: str, lambda_arns: List[str]
):
    """Creates Terraform resource for step functions

    Args:
        scope
        pipeline (Pipeline): pipeline to convert into step function
        aws_region (str): AWS Region
        lambda_arns (list): ARNs of Lambdas, required to give step functions access to invoke Lambdas
    """
    role = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": "sts:AssumeRole",
                "Principal": {
                    "Service": [
                        f"states.{aws_region}.amazonaws.com",
                        "events.amazonaws.com",
                    ]
                },
            }
        ],
    }
    policy = {
        "Version": "2012-10-17",
        "Statement": [
            {
                "Effect": "Allow",
                "Action": ["lambda:InvokeFunction"],
                "Resource": lambda_arns,
            },
            {
                "Effect": "Allow",
                "Action": [
                    "logs:CreateLogDelivery",
                    "logs:CreateLogStream",
                    "logs:GetLogDelivery",
                    "logs:UpdateLogDelivery",
                    "logs:DeleteLogDelivery",
                    "logs:ListLogDeliveries",
                    "logs:PutLogEvents",
                    "logs:PutResourcePolicy",
                    "logs:DescribeResourcePolicies",
                    "logs:DescribeLogGroups",
                ],
                "Resource": ["*"],
            },
            {
                "Effect": "Allow",
                "Action": [
                    "xray:PutTraceSegments",
                    "xray:PutTelemetryRecords",
                    "xray:GetSamplingRules",
                    "xray:GetSamplingTargets",
                ],
                "Resource": ["*"],
            },
        ],
    }
    log_group = cloudwatch_log_group.CloudwatchLogGroup(
        scope, f"{pipeline.name}log", name_prefix="/aws/vendedlogs/states/stepfunction"
    )
    stepfunction_policy = iam_policy.IamPolicy(
        scope,
        f"{pipeline.name}policy",
        policy=json.dumps(policy),
    )
    stepfunction_role = iam_role.IamRole(
        scope,
        f"{pipeline.name}role",
        assume_role_policy=json.dumps(role),
        name_prefix=pipeline.name,
    )
    stepfunction_policy_attachment = iam_role_policy_attachment.IamRolePolicyAttachment(
        scope,
        f"{pipeline.name}policyattachment",
        policy_arn=stepfunction_policy.arn,
        role=stepfunction_role.name,
    )
    step_function = sfn_state_machine.SfnStateMachine(
        scope,
        pipeline.name,
        role_arn=stepfunction_role.arn,
        name=pipeline.name,
        type="STANDARD",
        definition=json.dumps(pipeline.generate_step_functions()),
        logging_configuration={
            "include_execution_data": True,
            "level": "ALL",
            "log_destination": f"{log_group.arn}:*",
        },
        tracing_configuration={"enabled": True},
    )
    TerraformOutput(scope, f"{pipeline.name}_stepfunction_arn", value=step_function.arn)
    return step_function


class StepInLine(TerraformStack):
    def __init__(
        self,
        scope: Construct,
        name: str,
        pipeline: Pipeline,
        region: str,
        template_file: str = impresources.files(__package__) / "template_lambda.py",
        vpc_id: Optional[str] = None,
        subnet_filter: Optional[
            Union[IResolvable, List[data_aws_subnets.DataAwsSubnetsFilter]]
        ] = None,
        outbound_cidr: Optional[List[str]] = [
            "0.0.0.0/0",
        ],
    ):
        """Initialize a StepInLine terraform stack

        Args:
            scope (Construct)
            name (str): Unique name for Stack resource
            pipeline (Pipeline): The pipeline to instantiate in AWS
            region (str): AWS Region
            template_file (str): Location of template file to populate.  Defaults to internal template, but a custom file can be provided.
            vpc_id (Optional[str]): If Lambda needs to be in a VPC, supply the VPC ID
            subnet_filter: If vpc_id is needed, provide a filter to access the subnets
            outbound_cidr: Optional[List[str]]: The CIDRs to allow Lambda to access.  Only required if VPC is needed.
        """
        super().__init__(scope, name)

        AwsProvider(self, "AWS", region=region)

        subnet_ids = None
        security_group_ids = None
        if vpc_id is not None:
            subnets = data_aws_subnets.DataAwsSubnets(
                self, "private_subnets", filter=subnet_filter
            )
            security_group_for_lambda = security_group.SecurityGroup(
                self,
                "security_group_lambda",
                egress=[
                    {
                        "fromPort": 0,
                        "toPort": 0,
                        "protocol": "-1",
                        "cidrBlocks": outbound_cidr,
                    }
                ],
                vpc_id=vpc_id,
            )

            subnet_ids = subnets.ids
            security_group_ids = [security_group_for_lambda.id]
            logger.info(f"Successfully generated VPC Terraform resources")

        step_to_lambda_tf = {}
        for step in pipeline.get_steps():
            step_lambda = generate_lambda_function(
                self,
                pipeline.name,
                step,
                template_file,
                subnet_ids,
                security_group_ids,
            )
            step_to_lambda_tf[step.name] = step_lambda.arn
            logger.info(
                f"Successfully generated Lambda Terraform resource for step {step.name}"
            )

        pipeline.set_generate_step_name(lambda s: step_to_lambda_tf[s.name])
        step_function = generate_step_function(
            self, pipeline, region, list(step_to_lambda_tf.values())
        )
        logger.info(f"Successfully generated Step Function Terraform resource")
        if pipeline.schedule is not None:
            generate_event_bridge(self, pipeline, step_function.arn)
            logger.info(f"Successfully generated Event Bridge Terraform resource")


def rename_tf_output(path: Path):
    """Terraform creates .tf, but expects .tf.json.  This function
            adds the .json extension and moves it into the root
            directory

    Args:
        path (Path): directory where Terraform outputs the .tf file
    """
    for file in path.glob("*.tf"):
        new_name = file.with_suffix(".tf.json")
        file.replace(new_name.name)  # put in root directory


## example, for testing
def main():
    logging.basicConfig(level=logging.INFO)
    app = App(hcl_output=True)

    @step
    def preprocess(arg1: str) -> str:
        return "hello"

    @step
    def preprocess_2(arg1: str) -> str:
        return "hello"

    @step
    def preprocess_3(arg1: str) -> str:
        return "hello"

    @step
    def train(arg1: str, arg2: str, arg3: str):
        return "goodbye"

    step_process_result = preprocess("hi")
    step_process_result_2 = preprocess_2(step_process_result)
    step_process_result_3 = preprocess_3(step_process_result)
    step_train_result = train(
        step_process_result, step_process_result_2, step_process_result_3
    )
    instance_name = "aws_instance"
    pipe = Pipeline("mytest", steps=[step_train_result], schedule="rate(2 minutes)")
    print(pipe.generate_step_functions())
    stack = StepInLine(app, instance_name, pipe, "us-east-1")
    tf_path = Path(app.outdir, "stacks", instance_name)
    app.synth()
    rename_tf_output(tf_path)



from .fields import Field
from .inputs import Placeholder, StepInput
import json
import logging
from enum import Enum

logger = logging.getLogger(__name__)


def to_pascalcase(text):
    return "".join([t.title() for t in text.split("_")])


class Block(object):
    """
    Base class to abstract blocks used in `Amazon States Language <https://states-language.net/spec.html>`_.
    """

    def __init__(self, **kwargs):
        self.fields = kwargs
        for k, v in self.fields.items():
            if not self.is_field_allowed(k):
                raise TypeError("Field '{field}' is not supported.".format(field=k))

    def __getattr__(self, name):
        return self.fields.get(name, None)

    def is_field_allowed(self, field_name):
        return field_name in [field.value for field in self.allowed_fields()]

    def allowed_fields(self):
        return []

    def _replace_placeholders(self, params):
        if not isinstance(params, dict):
            return params
        modified_parameters = {}
        for k, v in params.items():
            if isinstance(v, Placeholder):
                modified_key = "{key}.$".format(key=k)
                modified_parameters[modified_key] = v.to_jsonpath()
            elif isinstance(v, dict):
                modified_parameters[k] = self._replace_placeholders(v)
            elif isinstance(v, list):
                modified_parameters[k] = [self._replace_placeholders(i) for i in v]
            else:
                modified_parameters[k] = v
        return modified_parameters

    def to_dict(self):
        result = {}
        fields_accepted_as_none = ("result_path", "input_path", "output_path")
        # Common fields
        for k, v in self.fields.items():
            if v is not None or k in fields_accepted_as_none:
                k = to_pascalcase(k)
                if k == to_pascalcase(Field.Parameters.value):
                    result[k] = self._replace_placeholders(v)
                else:
                    result[k] = v

        return result

    def to_json(self, pretty=False):
        """Serialize to a JSON formatted string.

        Args:
            pretty (bool, optional): Boolean flag set to `True` if JSON string should be prettified. `False`, otherwise. (default: False)

        Returns:
            str: JSON formatted string representation of the block.
        """
        if pretty:
            return json.dumps(self.to_dict(), indent=4)

        return json.dumps(self.to_dict())

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join(["{}={!r}".format(k, v) for k, v in self.fields.items()]),
        )

    def __str__(self):
        return self.to_json(pretty=True)


class Retry(Block):
    """
    A class for creating a Retry block.
    """

    def __init__(self, **kwargs):
        """Initialize a Retry block.

        Args:
            error_equals (list(str)): Non-empty list of strings, which match `Error Names <https://states-language.net/spec.html#error-names>`_. When a state reports an error, the interpreter scans through the retriers and, when the Error Name appears in the value of of a retrier’s `error_equals` field, implements the retry policy described in that retrier.
            interval_seconds (int, optional): Positive integer representing the number of seconds before the first retry attempt. (default: 1)
            max_attempts (int, optional): Non-negative integer representing the maximum number of retry attempts. (default: 3)
            backoff_rate(float, optional): A number which is the multiplier that increases the retry interval on each attempt. (default: 2.0)
        """
        super(Retry, self).__init__(**kwargs)

    def allowed_fields(self):
        return [
            Field.ErrorEquals,
            Field.IntervalSeconds,
            Field.MaxAttempts,
            Field.BackoffRate,
        ]


class State(Block):
    """
    Base class to abstract states in `Amazon States Language <https://states-language.net/spec.html>`_.
    """

    def __init__(self, state_id, state_type, output_schema=None, **kwargs):
        """
        Args:
            state_id (str): State name whose length **must be** less than or equal to 128 unicode characters. State names **must be** unique within the scope of the whole state machine.
            state_type (str): Type of the state. (Allowed values: `'Pass'`, `'Succeed'`, `'Fail'`, `'Wait'`, `'Task'`, `'Choice'`, `'Parallel'`, `'Map'`).
            output_schema (dict): Expected output schema for the State. This is used to validate placeholder inputs used by the next state in the state machine. (default: None)
            comment (str, optional): Human-readable comment or description. (default: None)
            input_path (str, optional): Path applied to the state’s raw input to select some or all of it; that selection is used by the state. (default: '$')
            parameters (dict, optional): The value of this field becomes the effective input for the state.
            result_path (str, optional): Path specifying the raw input’s combination with or replacement by the state’s result. (default: '$')
            output_path (str, optional): Path applied to the state’s output after the application of `result_path`, producing the effective output which serves as the raw input for the next state. (default: '$')
        """
        super(State, self).__init__(**kwargs)
        self.fields["type"] = state_type

        self.state_type = state_type
        self.state_id = state_id
        self.output_schema = output_schema
        self.step_output = StepInput(schema=output_schema)
        self.retries = []
        self.catches = []
        self.next_step = None
        self.in_chain = None

    def __repr__(self):
        return self.state_id + " " + super(State, self).__repr__()

    def allowed_fields(self):
        return [
            Field.Comment,
            Field.InputPath,
            Field.OutputPath,
            Field.Parameters,
            Field.ResultPath,
        ]

    def update_parameters(self, params):
        """
        Update `parameters` field in the state, if supported.

        Args:
            params (dict or list): The value of this field becomes the effective input for the state.
        """
        if Field.Parameters in self.allowed_fields():
            self.fields[Field.Parameters.value] = params

    def next(self, next_step):
        """
        Specify the next state or chain to transition to.

        Args:
            next_step (State or Chain): Next state or chain to transition to.

        Returns:
            State or Chain: Next state or chain that will be transitioned to.
        """
        if self.type in ("Succeed", "Fail"):
            raise ValueError(
                "Unexpected State instance `{step}`, State type `{state_type}` does not support method `next`.".format(
                    step=next_step, state_type=self.type
                )
            )

        # By design, Choice states do not have the Next field. When used in a chain, the subsequent step becomes the
        # default choice that executes if none of the specified rules match.
        # See language spec for more info: https://states-language.net/spec.html#choice-state
        if self.type == "Choice":
            if self.default is not None:
                logger.warning(
                    f"Chaining Choice state: Overwriting {self.state_id}'s current default_choice ({self.default.state_id}) with {next_step.state_id}"
                )
            self.default_choice(next_step)
            return self.default

        self.next_step = next_step
        return self.next_step

    def output(self):
        """
        Get the placeholder collection for the State's output.

        Returns:
            StepInput: Placeholder collection representing the State's output, and consequently the input to the next state in the workflow (if one exists).
        """
        return self.step_output

    def accept(self, visitor):
        if visitor.is_visited(self):
            return

        visitor.visit(self)
        if self.next_step is not None:
            self.next_step.accept(visitor)
        for catch in self.catches:
            catch.next_step.accept(visitor)

    def add_retry(self, retry):
        """
        Add a retrier or a list of retriers to the tail end of the list of retriers for the state.
        See `Error handling in Step Functions <https://docs.aws.amazon.com/step-functions/latest/dg/concepts-error-handling.html#error-handling-retrying-after-an-error>`_ for more details.

        Args:
            retry (Retry or list(Retry)): A retrier or list of retriers to add.
        """
        if Field.Retry in self.allowed_fields():
            (
                self.retries.extend(retry)
                if isinstance(retry, list)
                else self.retries.append(retry)
            )
        else:
            raise ValueError(
                f"{type(self).__name__} state does not support retry field. "
            )

    def add_catch(self, catch):
        """
        Add a catcher or a list of catchers to the tail end of the list of catchers for the state.
        See `Error handling in Step Functions <https://docs.aws.amazon.com/step-functions/latest/dg/concepts-error-handling.html#error-handling-fallback-states>`_ for more details.

        Args:
            catch (Catch or list(Catch): catcher or list of catchers to add.
        """
        if Field.Catch in self.allowed_fields():
            (
                self.catches.extend(catch)
                if isinstance(catch, list)
                else self.catches.append(catch)
            )
        else:
            raise ValueError(
                f"{type(self).__name__} state does not support catch field. "
            )

    def to_dict(self):
        result = super(State, self).to_dict()

        # Next step
        if self.next_step is not None:
            result[Field.Next.name] = self.next_step.state_id
        elif self.state_type not in ("Succeed", "Fail", "Choice"):
            result[Field.End.name] = True

        # Retry and catch
        if self.retries and self.is_field_allowed(Field.Retry.value):
            result[Field.Retry.name] = [retry.to_dict() for retry in self.retries]
        if self.catches and self.is_field_allowed(Field.Catch.value):
            result[Field.Catch.name] = [catch.to_dict() for catch in self.catches]

        return result


class Task(State):
    """
    Task State causes the interpreter to execute the work identified by the state’s `resource` field.
    """

    def __init__(self, state_id, retry=None, catch=None, **kwargs):
        """
        Args:
            state_id (str): State name whose length **must be** less than or equal to 128 unicode characters. State names **must be** unique within the scope of the whole state machine.
            retry (Retry or list(Retry), optional): A retrier or list of retriers that define the state's retry policy. See `Error handling in Step Functions <https://docs.aws.amazon.com/step-functions/latest/dg/concepts-error-handling.html#error-handling-retrying-after-an-error>`_ for more details.
            catch (Catch or list(Catch), optional): A catcher or list of catchers that define a fallback state. See `Error handling in Step Functions <https://docs.aws.amazon.com/step-functions/latest/dg/concepts-error-handling.html#error-handling-fallback-states>`_ for more details.
            resource (str): A URI that uniquely identifies the specific task to execute. The States language does not constrain the URI scheme nor any other part of the URI.
            timeout_seconds (int, optional): Positive integer specifying timeout for the state in seconds. If the state runs longer than the specified timeout, then the interpreter fails the state with a `States.Timeout` Error Name. (default: 60)
            timeout_seconds_path (str, optional): Path specifying the state's timeout value in seconds from the state input. When resolved, the path must select a field whose value is a positive integer.
            heartbeat_seconds (int, optional): Positive integer specifying heartbeat timeout for the state in seconds. This value should be lower than the one specified for `timeout_seconds`. If more time than the specified heartbeat elapses between heartbeats from the task, then the interpreter fails the state with a `States.Timeout` Error Name.
            heartbeat_seconds_path (str, optional): Path specifying the state's heartbeat value in seconds from the state input. When resolved, the path must select a field whose value is a positive integer.
            comment (str, optional): Human-readable comment or description. (default: None)
            input_path (str, optional): Path applied to the state’s raw input to select some or all of it; that selection is used by the state. (default: '$')
            parameters (dict, optional): The value of this field becomes the effective input for the state.
            result_path (str, optional): Path specifying the raw input’s combination with or replacement by the state’s result. (default: '$')
            output_path (str, optional): Path applied to the state’s output after the application of `result_path`, producing the effective output which serves as the raw input for the next state. (default: '$')
        """
        super(Task, self).__init__(state_id, "Task", **kwargs)
        if self.timeout_seconds is not None and self.timeout_seconds_path is not None:
            raise ValueError(
                "Only one of 'timeout_seconds' or 'timeout_seconds_path' can be provided."
            )

        if (
            self.heartbeat_seconds is not None
            and self.heartbeat_seconds_path is not None
        ):
            raise ValueError(
                "Only one of 'heartbeat_seconds' or 'heartbeat_seconds_path' can be provided."
            )

        if retry:
            self.add_retry(retry)

        if catch:
            self.add_catch(catch)

    def allowed_fields(self):
        return [
            Field.Comment,
            Field.InputPath,
            Field.OutputPath,
            Field.Parameters,
            Field.ResultPath,
            Field.TimeoutSeconds,
            Field.TimeoutSecondsPath,
            Field.HeartbeatSeconds,
            Field.HeartbeatSecondsPath,
            Field.Resource,
            Field.Retry,
            Field.Catch,
        ]


class DuplicateStatesInChain(Exception):
    pass


class Chain(object):
    """
    Chain is a logical group of states, that resembles a linked list. The start state is the head of the *Chain* and the end state is the tail of the *Chain*.
    """

    def __init__(self, steps=[]):
        """
        Args:
            steps (list(State), optional): List of states to be chained in-order. (default: [])
        """
        if not isinstance(steps, list):
            raise TypeError(
                "Chain takes a 'list' of steps. You provided an input that is not a list."
            )
        self.steps = []

        steps_expanded = []
        [
            (
                steps_expanded.extend(step)
                if isinstance(step, Chain)
                else steps_expanded.append(step)
            )
            for step in steps
        ]
        for step in steps_expanded:
            if steps_expanded.count(step) > 1:
                raise DuplicateStatesInChain("Duplicate states in the chain.")
        list(map(self.append, steps_expanded))

    def __iter__(self):
        return iter(self.steps)

    @property
    def state_id(self):
        if len(self.steps) == 0:
            raise ValueError("The chain is empty")
        return self.steps[0].state_id

    def append(self, step):
        """Add a state at the tail end of the chain.

        Args:
            step (State): State to insert at the tail end of the chain.
        """
        if len(self.steps) == 0:
            self.steps.append(step)
        else:
            if step in self.steps:
                raise DuplicateStatesInChain(
                    "State '{step_name}' is already inside this chain. A chain cannot have duplicate states.".format(
                        step_name=step.state_id
                    )
                )
            last_step = self.steps[-1]
            last_step.next(step)
            self.steps.append(step)

    def accept(self, visitor):
        for step in self.steps:
            step.accept(visitor)

    def __repr__(self):
        return "{}(steps={!r})".format(self.__class__.__name__, self.steps)


class GraphVisitor(object):

    def __init__(self):
        self.states = {}

    def is_visited(self, state):
        return state.state_id in self.states

    def visit(self, state):
        self.states[state.state_id] = state.to_dict()


class ValidationVisitor(object):

    def __init__(self):
        self.states = {}

    def is_visited(self, state):
        if self.states.get(state.state_id) == state.to_dict():
            return True
        else:
            return False

    def visit(self, state):
        if state.state_id in self.states:
            raise ValueError(
                "Each state in a workflow must have a unique state id. Found duplicate state id '{}' in workflow.".format(
                    state.state_id
                )
            )
        self.states[state.state_id] = state.to_dict()
        if state.next_step is None:
            return
        if (
            not hasattr(state.next_step, "fields")
            or Field.Parameters.value not in state.next_step.fields
        ):
            return
        params = state.next_step.fields[Field.Parameters.value]
        valid, invalid_param_name = self._validate_next_step_params(
            params, state.step_output
        )
        if not valid:
            raise ValueError(
                "State '{state_name}' is using an illegal placeholder for the '{param_name}' parameter.".format(
                    state_name=state.next_step.state_id, param_name=invalid_param_name
                )
            )

    def _validate_next_step_params(self, params, step_output):
        for k, v in params.items():
            if isinstance(v, StepInput):
                if v is not step_output and not step_output.contains(v):
                    return False, k
            elif isinstance(v, dict):
                valid, invalid_param_name = self._validate_next_step_params(
                    v, step_output
                )
                if not valid:
                    return valid, invalid_param_name
        return True, None


class Graph(Block):

    def __init__(self, branch, **kwargs):
        if not isinstance(branch, (State, Chain)):
            raise ValueError(
                "Expected branch to be a State or a Chain, but got `{branch}`".format(
                    branch=branch
                )
            )

        super(Graph, self).__init__(**kwargs)

        self.branch = branch
        self.states = {}
        self.build_graph(branch)

    def allowed_fields(self):
        return [Field.TimeoutSeconds, Field.Comment, Field.Version]

    def contains(self, state):
        return self.states.get(state.state_id, False)

    def build_graph(self, state):
        graph_visitor = GraphVisitor()
        validation_visitor = ValidationVisitor()
        state.accept(validation_visitor)
        state.accept(graph_visitor)
        self.states = graph_visitor.states

    def to_dict(self):
        result = super(Graph, self).to_dict()
        result["StartAt"] = self.branch.state_id
        result["States"] = self.states
        return result


class Parallel(State):
    """
    Parallel State causes parallel execution of "branches".

    A Parallel state causes the interpreter to execute each branch as concurrently as possible, and wait until each branch terminates (reaches a terminal state) before processing the next state in the Chain.
    """

    def __init__(self, state_id, retry=None, catch=None, **kwargs):
        """
        Args:
            state_id (str): State name whose length **must be** less than or equal to 128 unicode characters. State names **must be** unique within the scope of the whole state machine.
            retry (Retry or list(Retry), optional): A retrier or list of retriers that define the state's retry policy. See `Error handling in Step Functions <https://docs.aws.amazon.com/step-functions/latest/dg/concepts-error-handling.html#error-handling-retrying-after-an-error>`_ for more details.
            catch (Catch or list(Catch), optional): A catcher or list of catchers that define a fallback state. See `Error handling in Step Functions <https://docs.aws.amazon.com/step-functions/latest/dg/concepts-error-handling.html#error-handling-fallback-states>`_ for more details.
            comment (str, optional): Human-readable comment or description. (default: None)
            input_path (str, optional): Path applied to the state’s raw input to select some or all of it; that selection is used by the state. (default: '$')
            parameters (dict, optional): The value of this field becomes the effective input for the state.
            result_path (str, optional): Path specifying the raw input’s combination with or replacement by the state’s result. (default: '$')
            output_path (str, optional): Path applied to the state’s output after the application of `result_path`, producing the effective output which serves as the raw input for the next state. (default: '$')
        """
        super(Parallel, self).__init__(state_id, "Parallel", **kwargs)
        self.branches = []

        if retry:
            self.add_retry(retry)

        if catch:
            self.add_catch(catch)

    def allowed_fields(self):
        return [
            Field.Comment,
            Field.InputPath,
            Field.OutputPath,
            Field.Parameters,
            Field.ResultPath,
            Field.Retry,
            Field.Catch,
        ]

    def add_branch(self, branch):
        """
        Add a `State` or `Chain` as a branch to the Parallel state.

        Args:
            branch (State or Chain): State or Chain to attach as a branch to the Parallel state.
        """
        self.branches.append(branch)

    def to_dict(self):
        result = super(Parallel, self).to_dict()
        result[Field.Branches.name] = [
            Graph(branch).to_dict() for branch in self.branches
        ]
        return result


class Retry(Block):
    """
    A class for creating a Retry block.
    """

    def __init__(self, **kwargs):
        """Initialize a Retry block.

        Args:
            error_equals (list(str)): Non-empty list of strings, which match `Error Names <https://states-language.net/spec.html#error-names>`_. When a state reports an error, the interpreter scans through the retriers and, when the Error Name appears in the value of of a retrier’s `error_equals` field, implements the retry policy described in that retrier.
            interval_seconds (int, optional): Positive integer representing the number of seconds before the first retry attempt. (default: 1)
            max_attempts (int, optional): Non-negative integer representing the maximum number of retry attempts. (default: 3)
            backoff_rate(float, optional): A number which is the multiplier that increases the retry interval on each attempt. (default: 2.0)
        """
        super(Retry, self).__init__(**kwargs)

    def allowed_fields(self):
        return [
            Field.ErrorEquals,
            Field.IntervalSeconds,
            Field.MaxAttempts,
            Field.BackoffRate,
        ]


class IntegrationPattern(Enum):
    """
    Integration pattern enum classes for task integration resource arn builder
    """

    WaitForTaskToken = "waitForTaskToken"
    WaitForCompletion = "sync"
    CallAndContinue = ""


def get_service_integration_arn(
    service, api, integration_pattern=IntegrationPattern.CallAndContinue, version=None
):
    """
    ARN builder for task integration
    Args:
        service (str): The service name for the service integration
        api (str): The api of the service integration
        integration_pattern (IntegrationPattern, optional): The integration pattern for the task. (Default: IntegrationPattern.CallAndContinue)
        version (int, optional): The version of the resource to use. (Default: None)
    """
    arn = ""
    if integration_pattern == IntegrationPattern.CallAndContinue:
        arn = f"arn:aws:states:::{service}:{api.value}"
    else:
        arn = f"arn:aws:states:::{service}:{api.value}.{integration_pattern.value}"

    if version:
        arn = f"{arn}:{str(version)}"

    return arn


LAMBDA_SERVICE_NAME = "lambda"


class LambdaApi(Enum):
    Invoke = "invoke"


class LambdaStep(Task):
    """
    Creates a Task state to invoke an AWS Lambda function. See `Invoke Lambda with Step Functions <https://docs.aws.amazon.com/step-functions/latest/dg/connect-lambda.html>`_ for more details.
    """

    def __init__(self, state_id, wait_for_callback=False, **kwargs):
        """
        Args:
            state_id (str): State name whose length **must be** less than or equal to 128 unicode characters. State names **must be** unique within the scope of the whole state machine.
            wait_for_callback(bool, optional): Boolean value set to `True` if the Task state should wait for callback to resume the operation. (default: False)
            timeout_seconds (int, optional): Positive integer specifying timeout for the state in seconds. If the state runs longer than the specified timeout, then the interpreter fails the state with a `States.Timeout` Error Name. (default: 60)
            timeout_seconds_path (str, optional): Path specifying the state's timeout value in seconds from the state input. When resolved, the path must select a field whose value is a positive integer.
            heartbeat_seconds (int, optional): Positive integer specifying heartbeat timeout for the state in seconds. This value should be lower than the one specified for `timeout_seconds`. If more time than the specified heartbeat elapses between heartbeats from the task, then the interpreter fails the state with a `States.Timeout` Error Name.
            heartbeat_seconds_path (str, optional): Path specifying the state's heartbeat value in seconds from the state input. When resolved, the path must select a field whose value is a positive integer.
            comment (str, optional): Human-readable comment or description. (default: None)
            input_path (str, optional): Path applied to the state’s raw input to select some or all of it; that selection is used by the state. (default: '$')
            parameters (dict, optional): The value of this field becomes the effective input for the state.
            result_path (str, optional): Path specifying the raw input’s combination with or replacement by the state’s result. (default: '$')
            output_path (str, optional): Path applied to the state’s output after the application of `result_path`, producing the effective output which serves as the raw input for the next state. (default: '$')
        """

        if wait_for_callback:
            """
            Example resource arn: arn:aws:states:::lambda:invoke.waitForTaskToken
            """

            kwargs[Field.Resource.value] = get_service_integration_arn(
                LAMBDA_SERVICE_NAME,
                LambdaApi.Invoke,
                IntegrationPattern.WaitForTaskToken,
            )
        else:
            """
            Example resource arn: arn:aws:states:::lambda:invoke
            """

            kwargs[Field.Resource.value] = get_service_integration_arn(
                LAMBDA_SERVICE_NAME, LambdaApi.Invoke
            )

        super(LambdaStep, self).__init__(state_id, **kwargs)
